from typing import Dict, Optional

from airflow import AirflowException
from airflow.models import BaseOperator

from airflow_spell import SpellClient


""" https://spell.run/docs/runs/
"""


class SpellRunOperator(BaseOperator, SpellClient):
    """
    Execute a run on Spell Run (same args as spell.client.runs.RunService

    Args:
        spell_conn_id (str): Airflow connection id for spell
        spell_owner (str, optional): Spell owner (if different from user account)
        task_id (str, optional):
        params (dict, optional):
        (all commands below passed to SpellClient)
        command (str): the command to run
        machine_type (str, optional): the machine type for the run (default: CPU)
        workspace_id (int, optional): the workspace ID for code to include in the run (default: None)
        commit_hash (str, optional): a specific commit hash in the workspace corresponding to :obj:`workspace_id`
            for code to include in the run (default: None)
        commit_label (str, optional): a commit label for code to include in the run. Only applicable
            if this is a workflow run (i.e., the :py:attr:`~spell.client.SpellClient.active_workflow` of the
            client is set or a :obj:`workflow_id` is provided) (default: None). The value must correspond
            to one of the commit labels specified upon workflow creation using the ``--repo`` option.
            Only applicable if a workspace is specified.
        github_url (str, optional): a GitHub URL to a repository for code to include in the run. Not applicable
            when :obj:`workspace_id` or :obj:`commit_label` is specified.
        github_ref (str, optional): a reference to a commit, branch, or tag in the repository corresponding to
            :obj:`github_url` for code to include in the run (default: master)
        pip_packages (:obj:`list` of :obj:`str`, optional): pip dependencies (default: None).
            For example: ``["moviepy", "scikit-image"]``
        apt_packages (:obj:`list` of :obj:`str`, optional): apt dependencies (default: None).
            For example: ``["python-tk", "ffmpeg"]``
        requirements_file (str, optional): a path to a requirements file
        envvars (:obj:`dict` of :obj:`str` -> :obj:`str`, optional): name to value mapping of
            environment variables for the run (default: None).
            For example: ``{"VARIABLE" : "VALUE", "LANG" : "C.UTF-8"}``
        cwd (str, optional): the working directory within the repository in which to execute the command
            (default: None). Only applicable if a workspace is specified.
        tensorboard_directory(str, optional): the path where tensorboard files will be read from.
        distributed(int, optional): execute a distributed run using N machines of the specified machine type.
        conda_file (str, optional): the path to a conda specification file or YAML environment file (default: None).
        docker_image (str, optional): the name of docker image to use as base (default: None)
        framework (str, optional): the framework to use for the run (default: None). For example: ``pytorch``
        framework_version (str, optional): the framework version to use for the run (default: None).
            For example: ``0.2.0``
        attached_resources (:obj:`dict` of :obj:`str` -> :obj:`str`, optional): resource name to
            mountpoint mapping of attached resouces for the run (default: None).
            For example: ``{"runs/42" : "/mnt/data"}``
        description (str, optional): a description for the run (default: None)
        idempotent (bool, optional): use an existing identical run if available in lieu of re-running
            (default: false)
        workflow_id (int, optional): the id of the workflow to which this run will be associated (default: None).
            This argument is unnecessary if the :py:attr:`~spell.client.SpellClient.active_workflow` of
            the client is set and this argument will take precedence if they differ.
    """

    ui_color = "#f2f0f6"
    ui_fgcolor = "#3c1fd1"

    def __init__(
        self,
        *,
        task_id: str,
        spell_owner: Optional[str] = None,
        spell_conn_id: Optional[str] = None,
        **kwargs,
    ):
        BaseOperator.__init__(self, task_id=task_id)
        SpellClient.__init__(self, spell_conn_id=spell_conn_id, spell_owner=spell_owner)
        self.log.warning(kwargs)
        if "default_args" in kwargs:
            kwargs.pop("default_args")
        self.kwargs = kwargs

    def execute(self, context: Dict) -> int:
        """
        Submit and monitor a Spell run
        :raises: AirflowException
        """
        self.submit_run(context)
        self.monitor_run(context)

        # this return value gets pushed as XCom
        return self.spell_run_id

    def submit_run(self, context: Dict):  # pylint: disable=unused-argument
        self.log.info("Running Spell run")

        try:
            run = self.client.runs.new(**self.kwargs)
            self.spell_run_id = run.id

            self.log.info(
                "Spell run (spell_run_id: %s) started: %s"
                % (self.spell_run_id, str(run))
            )

        except Exception as e:
            self.log.info("Spell run (task_id: %s) failed submission" % self.task_id)
            raise AirflowException(e)

    def monitor_run(self, context: Dict):  # pylint: disable=unused-argument
        """
        Monitor a Spell run
        :raises: AirflowException
        """
        try:
            self.wait_for_run(self.spell_run_id)
            self.check_run_complete(self.spell_run_id)
            self.log.info("Spell run (%s) succeeded" % self.spell_run_id)

        except Exception as e:
            self.log.info("Spell run (%s) failed monitoring" % self.spell_run_id)
            raise AirflowException(e)



from random import uniform
from time import sleep
from typing import List, Optional, Union

from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from airflow.utils.log.logging_mixin import LoggingMixin
from spell.client import SpellClient as ExternalSpellClient
from spell.client.runs import Run as ExternalSpellRun
from spell.client.runs import RunsService as ExternalSpellRunsService
from spell.cli.commands.ps import status_names


class SpellHook(BaseHook):
    def __init__(self, spell_conn_id="spell_default", owner: Optional[str] = None):
        super().__init__()
        self.spell_conn_id = spell_conn_id
        self.owner = owner

    def get_client(self):
        if self.owner is not None:
            owner = self.owner
        else:
            owner = self._get_owner()

        return ExternalSpellClient(token=self._get_token(), owner=owner)

    def _get_token(self):
        # get_connection is on BaseHook
        connection_object = self.get_connection(self.spell_conn_id)
        return connection_object.password

    def _get_owner(self):
        connection_object = self.get_connection(self.spell_conn_id)
        return connection_object.host


STILL_RUNNING = [
    ExternalSpellRunsService.BUILDING,
    ExternalSpellRunsService.PUSHING,
    ExternalSpellRunsService.RUNNING,
    ExternalSpellRunsService.SAVING,
]


class SpellClient(LoggingMixin):
    MAX_RETRIES = 4200
    STATUS_RETRIES = 10

    # delays are in seconds
    DEFAULT_DELAY_MIN = 1
    DEFAULT_DELAY_MAX = 10

    def __init__(
        self, spell_conn_id: Optional[str] = None, spell_owner: Optional[str] = None
    ):
        super().__init__()
        self.spell_conn_id = spell_conn_id
        self.spell_owner = spell_owner
        self._hook: Optional[SpellHook] = None
        self._client: Optional[ExternalSpellClient] = None

    @property
    def hook(self) -> SpellHook:
        if self._hook is None:
            self._hook = SpellHook(
                spell_conn_id=self.spell_conn_id, owner=self.spell_owner
            )
        return self._hook

    @property
    def client(self) -> ExternalSpellClient:
        if self._client is None:
            self._client = self.hook.get_client()
        return self._client

    def wait_for_run(self, run_id: str, delay: Optional[Union[int, float]] = None):
        """
        Wait for spell run to complete

        :param run_id: a spell run ID
        :type run_id: str

        :param delay: a delay before polling for run status
        :type delay: Optional[Union[int, float]]

        :raises: AirflowException
        """
        _delay(delay)
        self._poll_for_run_running(run_id, delay)
        self._poll_for_run_complete(run_id, delay)
        self.log.info("Spell run (%s) has completed" % run_id)

    def check_run_complete(self, run_id: str) -> bool:
        """
        Check the final status of the spell run; return True if the run
        'COMPLETE', else raise an AirflowException

        :param run_id: a spell run ID
        :type run_id: str

        :rtype: bool

        :raises: AirflowException
        """
        run: ExternalSpellRun = self._get_run(run_id)
        run_status = run.status

        if run_status == ExternalSpellRunsService.COMPLETE:
            if int(run.user_exit_code) == 0:
                self.log.info("Spell run (%s) completed: %s" % (run_id, run))
                return True
            else:
                raise AirflowException(
                    "Spell run (%s) completed with a non-zero exit code: %s"
                    % (run_id, run)
                )

        if run_status == ExternalSpellRunsService.FAILED:
            raise AirflowException("Spell run (%s) failed: %s" % (run_id, run))

        if run_status in STILL_RUNNING:
            raise AirflowException("Spell (%s) is not complete: %s" % (run_id, run))

        raise AirflowException(
            "Spell (%s) has unknown status (%s): %s" % (run_id, run_status, run)
        )

    def _get_run(self, run_id: str) -> ExternalSpellRun:
        return ExternalSpellRun(self.client.api, self.client.api.get_run(run_id))

    def _poll_for_run_running(self, run_id: str, delay: Union[int, float, None] = None):
        """
        Poll for job running. The status that indicates a job is running or
        already complete are: 'RUNNING'|'COMPLETED'|'FAILED'.

        So the status options that this will wait for are the transitions from:
        'SUBMITTED'>'PENDING'>'RUNNABLE'>'STARTING'>'RUNNING'|'COMPLETED'|'FAILED'

        The completed status options are included for cases where the status
        changes too quickly for polling to detect a RUNNING status that moves
        quickly from STARTING to RUNNING to completed (often a failure).

        :param run_id: a spell run ID
        :type run_id: str

        :param delay: a delay before polling for job status
        :type delay: Optional[Union[int, float]]

        :raises: AirflowException
        """
        _delay(delay)
        running_status = [
            ExternalSpellRunsService.BUILDING,
            ExternalSpellRunsService.RUNNING,
            ExternalSpellRunsService.SAVING,
            ExternalSpellRunsService.PUSHING,
        ] + list(status_names.keys())
        self._poll_run_status(run_id, running_status)

    def _poll_for_run_complete(
        self, run_id: str, delay: Union[int, float, None] = None
    ):
        """
        Poll for job completion. The status that indicates job completion
        are: 'COMPLETED'|'FAILED'.

        So the status options that this will wait for are the transitions from:
        'SUBMITTED'>'PENDING'>'RUNNABLE'>'STARTING'>'RUNNING'>'COMPLETED'|'FAILED'

        :param run_id: a spell run ID
        :type run_id: str

        :param delay: a delay before polling for job status
        :type delay: Optional[Union[int, float]]

        :raises: AirflowException
        """
        _delay(delay)
        complete_status = ExternalSpellRunsService.FINAL
        self._poll_run_status(run_id, complete_status)

    def _poll_run_status(self, run_id: str, match_status: List[str]) -> bool:
        """
        Poll for job status using an exponential back-off strategy (with max_retries).

        :param run_id: a spell ID
        :type run_id: str

        :param match_status: a list of job status to match; the batch job status are:
            'SUBMITTED'|'PENDING'|'RUNNABLE'|'STARTING'|'RUNNING'|'COMPLETED'|'FAILED'
        :type match_status: List[str]

        :rtype: bool

        :raises: AirflowException
        """
        retries = 0
        while True:

            run = self._get_run(run_id)
            run_status = run.status

            self.log.debug(
                "Spell run (%s) check status (%s) in %s"
                % (run_id, run_status, match_status)
            )

            if run_status in match_status:
                return True

            if retries >= self.MAX_RETRIES:
                raise AirflowException(
                    "Spell run (%s) status checks exceed max_retries" % run_id
                )

            retries += 1
            pause = _exponential_delay(retries)

            self.log.info(
                "Spell run (%s) current status (%s), next check (%d of %d)"
                " in the %.2f seconds"
                % (run_id, run_status, retries, self.MAX_RETRIES, pause)
            )

            _delay(pause)


def _delay(delay: Union[int, float, None] = None):
    """
    Pause execution for ``delay`` seconds.

    :param delay: a delay to pause execution using ``time.sleep(delay)``;
        a small 1 second jitter is applied to the delay.
    :type delay: Optional[Union[int, float]]

    .. note::
        This method uses a default random delay, i.e.
        ``random.uniform(DEFAULT_DELAY_MIN, DEFAULT_DELAY_MAX)``;
        using a random interval helps to avoid AWS API throttle limits
        when many concurrent tasks request job-descriptions.
    """
    if delay is None:
        delay = uniform(SpellClient.DEFAULT_DELAY_MIN, SpellClient.DEFAULT_DELAY_MAX)
    else:
        delay = _add_jitter(delay)
    sleep(delay)


def _exponential_delay(tries: int) -> float:
    max_interval = 600.0  # results in 3 to 10 minute delay
    delay = 1 + pow(tries * 0.6, 2)
    delay = min(max_interval, delay)
    return uniform(delay / 3, delay)


def _add_jitter(
    delay: Union[int, float],
    width: Union[int, float] = 1,
    minima: Union[int, float] = 0,
) -> float:
    """
    Use delay +/- width for random jitter

    Adding jitter to status polling can help to avoid
    Spell API limits for monitoring spell jobs with
    a high concurrency in Airflow tasks.

    :param delay: number of seconds to pause;
        delay is assumed to be a positive number
    :type delay: Union[int, float]

    :param width: delay +/- width for random jitter;
        width is assumed to be a positive number
    :type width: Union[int, float]

    :param minima: minimum delay allowed;
        minima is assumed to be a non-negative number
    :type minima: Union[int, float]

    :return: uniform(delay - width, delay + width) jitter
        and it is a non-negative number
    :rtype: float
    """
    delay = abs(delay)
    width = abs(width)
    minima = abs(minima)
    lower = max(minima, delay - width)
    upper = delay + width
    return uniform(lower, upper)



from setuptools import setup, find_packages

setup(
    name="airflow-spell",
    version="0.0.9",
    author="Dan O'Donovan",
    author_email="dan.odonovan@healx.io",
    description="Apache Airflow integration for spell.run",
    package_dir={"": "src"},
    packages=find_packages(where="src", include=["airflow_spell", "airflow_spell.*"]),
    install_requires=[
        "apache-airflow>=2.0,<3.0",
        "spell>=0.38.4,<1.0",
    ],
)
